﻿using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Networks.IntegratedOptimization.SurrogateModeling {
  [Item("Assisted Optimizer Problem", "")]
  [StorableClass]
  public sealed class AssistedOptimizerProblem : SingleObjectiveBasicProblem<RealVectorEncoding> {
    private const string RegressionSolutionParameterName = "RegressionSolution";
    private const string OptimizeExpectedImprovementParameterName = "OptimizeExpectedImprovement";

    public IValueParameter<IRegressionSolution> RegressionSolutionParameter {
      get { return (IValueParameter<IRegressionSolution>)Parameters[RegressionSolutionParameterName]; }
    }

    public IValueParameter<BoolValue> OptimizeExpectedImprovementParameter {
      get { return (IValueParameter<BoolValue>)Parameters[OptimizeExpectedImprovementParameterName]; }
    }

    public IRegressionSolution RegressionSolution {
      get { return RegressionSolutionParameter.Value; }
      set { RegressionSolutionParameter.Value = value; }
    }

    public BoolValue OptimizeExpectedImprovement {
      get { return OptimizeExpectedImprovementParameter.Value; }
      set { OptimizeExpectedImprovementParameter.Value = value; }
    }

    public Action<RealVector, double> NotifyEvaluation { get; set; }

    public override bool Maximization {
      get { return false; }
    }

    #region Constructors & Cloning
    [StorableConstructor]
    private AssistedOptimizerProblem(bool deserializing) : base(deserializing) { }
    private AssistedOptimizerProblem(AssistedOptimizerProblem original, Cloner cloner) : base(original, cloner) { }
    public AssistedOptimizerProblem() {
      Parameters.Add(new ValueParameter<IRegressionSolution>(RegressionSolutionParameterName, ""));
      Parameters.Add(new ValueParameter<BoolValue>(OptimizeExpectedImprovementParameterName, "", new BoolValue(true)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AssistedOptimizerProblem(this, cloner);
    }
    #endregion

    public override double Evaluate(Individual individual, IRandom random) {
      var point = individual.RealVector();
      var solution = RegressionSolutionParameter.Value;
      var useExpectedImprovement = OptimizeExpectedImprovement.Value;

      var result = ExpectedImprovementHelpers.Evaluate(new[] { point }, solution, useExpectedImprovement).Single();
      var target = useExpectedImprovement ? -result : result;

      NotifyEvaluation((RealVector)point.Clone(), target);

      return target;
    }

    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      base.Analyze(individuals, qualities, results, random);

      var bestIndividuals = individuals.Zip(qualities, (i, q) => new { Individual = i, Quality = q }).MinItems(x => x.Quality);
      var best = bestIndividuals.First();

      IResult bestSolutionResult;
      if (!results.TryGetValue("BestSolution", out bestSolutionResult)) {
        results.Add(bestSolutionResult = new Result("BestSolution", typeof(RealVector)));
      } else bestSolutionResult = results["BestSolution"];

      IResult bestQualityResult;
      if (!results.TryGetValue("BestQuality", out bestQualityResult)) {
        results.Add(bestQualityResult = new Result("BestQuality", typeof(DoubleValue)));
      } else bestQualityResult = results["BestQuality"];

      var bestQuality = (DoubleValue)bestQualityResult.Value;
      if (bestQuality == null || best.Quality <= bestQuality.Value) {
        bestSolutionResult.Value = (RealVector)best.Individual.RealVector().Clone();
        bestQualityResult.Value = new DoubleValue(best.Quality);
      }
    }
  }
}
