#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.Drawing;

namespace HeuristicLab.Core.Networks {
  [Item("MessageValue", "A value of a message sent between ports.")]
  [StorableClass]
  public class MessageValue<T> : Item, IMessageValue<T> where T : class, IItem {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Field; }
    }
    public override Image ItemImage {
      get {
        if (value != null) return value.ItemImage;
        else return base.ItemImage;
      }
    }

    [Storable]
    protected string name;
    public string Name {
      get { return name; }
    }
    public Type DataType {
      get { return typeof(T); }
    }
    [Storable]
    protected T value;
    public T Value {
      get { return value; }
      set {
        if (value != this.value) {
          DeregisterValueEvents();
          this.value = value;
          RegisterValueEvents();
          OnValueChanged();
        }
      }
    }
    IItem IMessageValue.Value {
      get { return Value; }
      set {
        T val = value as T;
        if ((value != null) && (val == null))
          throw new InvalidOperationException(
            string.Format("Type mismatch. Value is not a \"{0}\".",
                          typeof(T).GetPrettyName())
          );
        Value = val;
      }
    }

    [StorableConstructor]
    protected MessageValue(bool deserializing) : base(deserializing) { }
    protected MessageValue(MessageValue<T> original, Cloner cloner)
      : base(original, cloner) {
      this.name = original.name;
      this.value = cloner.Clone(original.value);
      RegisterValueEvents();
    }
    public MessageValue() {
      name = "MessageValue";
    }
    public MessageValue(string name) {
      if (name == null) this.name = string.Empty;
      else this.name = name;
    }
    public MessageValue(string name, T value)
      : this(name) {
      this.value = value;
      RegisterValueEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterValueEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MessageValue<T>(this, cloner);
    }

    public override string ToString() {
      if (Value == null)
        return string.Format("{0}: null", Name);
      else
        return string.Format("{0}: {1}", Name, Value.ToString());
    }

    public event EventHandler ValueChanged;
    protected virtual void OnValueChanged() {
      var handler = ValueChanged;
      if (handler != null) handler(this, EventArgs.Empty);
      OnItemImageChanged();
      OnToStringChanged();
    }
    protected virtual void RegisterValueEvents() {
      if (value != null) {
        value.ItemImageChanged += Value_ItemImageChanged;
        value.ToStringChanged += Value_ToStringChanged;
      }
    }
    protected virtual void DeregisterValueEvents() {
      if (value != null) {
        value.ItemImageChanged -= Value_ItemImageChanged;
        value.ToStringChanged -= Value_ToStringChanged;
      }
    }
    private void Value_ItemImageChanged(object sender, EventArgs e) {
      OnItemImageChanged();
    }
    private void Value_ToStringChanged(object sender, EventArgs e) {
      OnToStringChanged();
    }
  }
}
